/*
Copyright 2007-2024. Algoryx Simulation AB.

All AGX source code, intellectual property, documentation, sample code,
tutorials, scene files and technical white papers, are copyrighted, proprietary
and confidential material of Algoryx Simulation AB. You may not download, read,
store, distribute, publish, copy or otherwise disseminate, use or expose this
material unless having a written signed agreement with Algoryx Simulation AB, or having been
advised so by Algoryx Simulation AB for a time limited evaluation, or having purchased a
valid commercial license from Algoryx Simulation AB.

Algoryx Simulation AB disclaims all responsibilities for loss or damage caused
from using this software, unless otherwise stated in written agreements with
Algoryx Simulation AB.
*/

#pragma once

#include <agxSensor/LidarRayPatternGenerator.h>

namespace agxSensor
{
  /**
  Horizontal sweep pattern containing a full revolution of transforms, given
  field of view and resolution. The frequency, number of revolutions per second,
  affects the number of active rays during the time step size of the environment.
  */
  class AGXSENSOR_EXPORT LidarRayPatternHorizontalSweep : public LidarRayPatternGenerator
  {
  public:
    /**
    Default constructor, zero rays.
    */
    LidarRayPatternHorizontalSweep();

    /**
    Construct given field of view \p fov (radians x and y centered on middle of sphere), resolution (radians x
    and y) \p resolution and rotation frequency \p frequency (Hz).
    \param fov - field of view in radians x (horizontal) and y (vertical)
    \param resolution - resolution in radians x (horizontal) and y (vertical)
    \param frequency - number of revolutions per second
    */
    LidarRayPatternHorizontalSweep(const agx::Vec2& fov,
                                   const agx::Vec2& resolution,
                                   agx::Real frequency);

    /**
    Construct given field of view \p horizontal and vertical fov (radians x and y), resolution (radians x
    and y) \p resolution and rotation frequency \p frequency (Hz).
    \param horizontalFov - horizontal field of view in radians x (horizontal) and y (vertical) (default full rotation)
    \param verticalFov - veritcal field of view in radians x (horizontal) and y (vertical)
    \param resolution - resolution in radians x (horizontal) and y (vertical)
    \param frequency - number of revolutions per second
    */
    LidarRayPatternHorizontalSweep(const agx::RangeReal& horizontalFov,
                                   const agx::RangeReal& verticalFov,
                                   const agx::Vec2& resolution,
                                   agx::Real frequency);


    DOXYGEN_START_INTERNAL_BLOCK()

  public:
    virtual ~LidarRayPatternHorizontalSweep() = default;
    virtual LidarRayPatternInterval getNextInterval(agx::Real dt) override;
    virtual size_t getNumRays() const override;

    AGXSTREAM_DECLARE_SERIALIZABLE( agxSensor::LidarRayPatternHorizontalSweep );

  private:
    struct PatternState
    {
      agx::Real elapsedTime{0.0};
      agx::Real sweepTime{0.0};
      size_t last{std::numeric_limits<size_t>::max()};
    };

  private:
    void setRays(const agx::RangeReal& horizontalFov, const agx::RangeReal& verticalFov, const agx::Vec2& resolution, agx::Real frequency);

    DOXYGEN_END_INTERNAL_BLOCK()

  private:
    PatternState m_state;
    std::shared_ptr<LidarRayPatternStorage> m_pattern;
  };
}
